import os
import shutil
import tempfile

import pytest
from gramps.cli.clidbman import CLIDbManager
from gramps.gen.db import DbReadBase, DbTxn
from gramps.gen.db.utils import make_database
from gramps.gen.dbstate import DbState
from gramps.gen.lib import Person

from gramps_ql.gql import GQLQuery


@pytest.fixture
def db():
    """Return Gramps Database."""
    TEST_GRAMPSHOME = tempfile.mkdtemp()
    os.environ["GRAMPSHOME"] = TEST_GRAMPSHOME
    dbman = CLIDbManager(DbState())
    path, name = dbman.create_new_db_cli("GQL Test", dbid="sqlite")
    db = make_database("sqlite")
    db.load(path)
    person = Person()
    person.gramps_id = "person001"
    with DbTxn("Add test objects", db) as trans:
        db.add_person(person, trans)
    yield db
    db.close()
    shutil.rmtree(TEST_GRAMPSHOME)


def test_fixture(db):
    assert isinstance(db, DbReadBase)


def test_person_gramps_id(db):
    q = GQLQuery("type=person")
    assert len(list(q.iter_objects(db))) == 1
    for obj in q.iter_objects(db):
        assert isinstance(obj, Person)
    q = GQLQuery("""type=person and gramps_id="person001" """)
    assert len(list(q.iter_objects(db))) == 1
    q = GQLQuery("""type=person and gramps_id!="person001" """)
    assert len(list(q.iter_objects(db))) == 0
    q = GQLQuery("""type=person and gramps_id="person002" """)
    assert len(list(q.iter_objects(db))) == 0
    q = GQLQuery("""type=person and gramps_id>"person002" """)
    assert len(list(q.iter_objects(db))) == 0
    q = GQLQuery("""type=person and gramps_id<"person002" """)
    assert len(list(q.iter_objects(db))) == 1
    q = GQLQuery("type=person and gramps_id < 'person002'")
    assert len(list(q.iter_objects(db))) == 1
    q = GQLQuery("type=person and gramps_id < 'person002'")
    assert len(list(q.iter_objects(db))) == 1



from __future__ import annotations  # can be removed at 3.8 EOL

__version__: str | None
__version_tuple__: tuple | None

try:
    # This file is auto-generated, and could be missing.
    # Tell mypy to ignore it
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:
    __version__ = None
    __version_tuple__ = None

__all__ = (
    "__version__",
    "__version_tuple__",
)

from .gql import iter_objects, match, parse



from gramps_ql import parse


def test_single():
    assert parse("type=person").as_list() == ["type", "=", "person"]


def test_two():
    assert parse("type=person or date.year > 2021").as_list() == [
        [
            "type",
            "=",
            "person",
            "or",
            "date.year",
            ">",
            "2021",
        ]
    ]


def test_three():
    assert parse("type=person or name='John Doe' and date.year > 2021").as_list() == [
        [
            "type",
            "=",
            "person",
            "or",
            ["name", "=", "'John Doe'", "and", "date.year", ">", "2021"],
        ]
    ]
    assert parse("type=person and name='John Doe' or date.year > 2021").as_list() == [
        [
            ["type", "=", "person", "and", "name", "=", "'John Doe'"],
            "or",
            "date.year",
            ">",
            "2021",
        ]
    ]
    assert parse("type=person and name='John Doe' or only_id").as_list() == [
        [
            ["type", "=", "person", "and", "name", "=", "'John Doe'"],
            "or",
            "only_id",
        ]
    ]


def test_brackets():
    assert parse("(type=person or name='John Doe') and date.year > 2021").as_list() == [
        [
            ["type", "=", "person", "or", "name", "=", "'John Doe'"],
            "and",
            "date.year",
            ">",
            "2021",
        ]
    ]
    assert parse(
        "(((((((((((type=person or name='John Doe')))) and date.year > 2021)))))))"
    ).as_list() == [
        [
            ["type", "=", "person", "or", "name", "=", "'John Doe'"],
            "and",
            "date.year",
            ">",
            "2021",
        ]
    ]



from gramps_ql import match
from gramps_ql.gql import GQLQuery


def test_match_single_string():
    q = GQLQuery("test")
    assert q._match_single({"one": "rhs"}, "one", "=", "rhs")
    assert not q._match_single({"one": "x"}, "one", "=", "rhs")
    assert q._match_single({"one": {"two": "rhs"}}, "one.two", "=", "rhs")
    assert not q._match_single({"one": {"x": "rhs"}}, "one.two", "=", "rhs")
    assert q._match_single({"one": {"two": ["rhs"]}}, "one.two[0]", "=", "rhs")
    assert not q._match_single({"one": {"two": ["rhs"]}}, "one.two[1]", "=", "rhs")
    assert q._match_single(
        {"one": {"two": [{"three": "rhs"}]}}, "one.two[0].three", "=", "rhs"
    )


def test_match_single_number():
    q = GQLQuery("test")
    assert q._match_single({"one": -1}, "one", "<", 0)
    assert not q._match_single({"one": "x"}, "one", "<", 0)
    assert q._match_single({"one": {"two": -1}}, "one.two", "<", 0)
    assert not q._match_single({"one": {"x": -1}}, "one.two", "<", 0)
    assert q._match_single({"one": {"two": [-1]}}, "one.two[0]", "<", 0)
    assert not q._match_single({"one": {"two": [-1]}}, "one.two[1]", "<", 0)
    assert q._match_single(
        {"one": {"two": [{"three": -1}]}}, "one.two[0].three", "<", 0
    )


def test_match():
    query = "one.two = x"
    q = GQLQuery(query)
    obj = {"one": {"two": "x"}, "three": {"four": ["y"]}}
    assert q.match(obj)
    assert match(query, obj)


def test_match_noop():
    q = GQLQuery("one.two")
    assert q.match({"one": {"two": 1}, "three": {"four": ["y"]}})


def test_match_and():
    q = GQLQuery("one.two = x and three.four[0] = y")
    assert q.match({"one": {"two": "x"}, "three": {"four": ["y"]}})


def test_match_and_or():
    q = GQLQuery("one.two = x and three.four[0] = y or five")
    assert q.match({"one": {"two": "x"}, "three": {"four": ["y"]}, "five": 1})


def test_length():
    q = GQLQuery("array.length = 1")
    assert not q.match({"array": []})
    assert not q.match({"array": [1, 1]})
    assert q.match({"array": [1]})


def test_array_contains():
    q = GQLQuery("array ~ 2")
    assert not q.match({"array": []})
    assert not q.match({"array": [3, 4, 5]})
    assert q.match({"array": [1, 2, 3]})


def test_string_contains():
    q = GQLQuery("string ~ 2")
    assert not q.match({"string": "abc"})
    assert not q.match({"string": "145"})
    assert q.match({"string": "co2"})


def test_any():
    q = GQLQuery("array.any = 2")
    assert not q.match({"array": []})
    assert not q.match({"array": [3, 4, 5]})
    assert q.match({"array": [1, 2, 3]})


def test_any_string():
    q = GQLQuery("array.any.value = a")
    assert not q.match({"array": []})
    assert not q.match({"array": [{"value": "c"}]})
    assert q.match({"array": [{"value": "b"}, {"value": "a"}]})


def test_any_number():
    q = GQLQuery("array.any.value = 3")
    assert not q.match({"array": []})
    assert not q.match({"array": [{"value": 1}]})
    assert q.match({"array": [{"value": 2}, {"value": 3}]})


def test_all():
    q = GQLQuery("array.all = 2")
    assert not q.match({"array": []})
    assert not q.match({"array": [2, 4, 5]})
    assert q.match({"array": [2, 2, 2]})


def test_all_string():
    q = GQLQuery("array.all.value = a")
    assert not q.match({"array": []})
    assert not q.match({"array": [{"value": "a"}, {"value": "b"}]})
    assert q.match({"array": [{"value": "a"}, {"value": "a"}]})


def test_all_number():
    q = GQLQuery("array.all.value = 3")
    assert not q.match({"array": []})
    assert not q.match({"array": [{"value": 1}, {"value": 3}]})
    assert q.match({"array": [{"value": 3}, {"value": 3}]})



"""Gramps Query Language."""

from __future__ import annotations  # can be removed at 3.8 EOL

import json
from collections.abc import Generator
from typing import Any, Optional

import pyparsing as pp
from gramps.gen.db import DbReadBase
from gramps.gen.lib import PrimaryObject
from gramps.gen.lib.serialize import to_json

pp.ParserElement.enablePackrat()


def match(query: str, obj: dict[str, Any]) -> bool:
    """Match a single object (given as dictionary) to a query."""
    gq = GQLQuery(query=query)
    return gq.match(obj)


def iter_objects(query: str, db: DbReadBase) -> Generator[PrimaryObject, None, None]:
    """Iterate over primary objects in a Gramps database."""
    gq = GQLQuery(query=query)
    return gq.iter_objects(db)


word = pp.Word(pp.alphanums + "." + "_")
rhs = word | pp.dbl_quoted_string | pp.sgl_quoted_string
# lhs = pp.Word(pp.identchars, pp.identbodychars + "." + "_")
logical_and = pp.CaselessKeyword("and")
logical_or = pp.CaselessKeyword("or")
logical = logical_and | logical_or
operator = pp.one_of("= != < <= > >= ~ !~")


property_name = pp.Word(pp.identchars, pp.identbodychars + "_")
attribute = "." + property_name
index = "[" + pp.common.integer + "]"
lhs = property_name + (attribute | index) * ...

expression = pp.Combine(lhs("lhs*")) + (operator("operator*") + rhs("rhs*")) * ...

infix = pp.infix_notation(
    expression,
    [
        (logical_and, 2, pp.OpAssoc.LEFT),
        (logical_or, 2, pp.OpAssoc.LEFT),
    ],
)

GRAMPS_OBJECT_NAMES = {
    "person": "people",
    "family": "families",
    "event": "events",
    "place": "places",
    "citation": "citations",
    "source": "sources",
    "repository": "repositories",
    "media": "media",
    "note": "notes",
}


def parse(query: str):
    """Parse a query."""
    return infix.parse_string(query, parse_all=True)


def parse_lhs(query: str):
    return lhs.parse_string(query, parse_all=True)


class GQLQuery:
    """GQL query class."""

    def __init__(self, query: str):
        self.query = query
        self.parsed = parse(self.query)

    @staticmethod
    def _combine_logical(op: Optional[str], value1: bool, value2: bool) -> bool:
        """Combine booleans with a logical operator as string."""
        if op == "and":
            return value1 and value2
        elif op == "or":
            return value1 or value2
        # if op is None, return the second value
        return value2

    def _traverse(self, parsed_list: list, obj: dict[str, Any]):
        """Traverse a nested parsed list."""
        result = None
        op = None
        expression = []
        for i, item in enumerate(parsed_list):
            if isinstance(item, list) or item in ["and", "or"]:
                if expression:
                    value = self._match_single(obj, *expression)
                    result = self._combine_logical(op, result, value)
                    op = None
                expression = []
                if item in ["and", "or"]:
                    op = item
                elif isinstance(item, list):
                    value = self._traverse(item, obj)
                    result = self._combine_logical(op, result, value)
                    op = None
            else:
                expression.append(item)
        if expression:
            value = self._match_single(obj, *expression)
            result = self._combine_logical(op, result, value)
        return result

    def match(self, obj: dict[str, Any]) -> bool:
        """Match an object to the query."""
        parsed_list = self.parsed.as_list()
        return self._traverse(parsed_list, obj)

    def _match_single(
        self, obj: dict[str, Any], lhs: str, operator: str = "", rhs: str = ""
    ) -> bool:
        """Match an object to a single condition."""
        parsed = parse_lhs(lhs)
        if not parsed:
            raise ValueError(f"Unable to parse left-hand side: {lhs}")
        for i, part in enumerate(parsed):
            if i == 0:
                result = obj.get(parsed[0])
            elif part in ["[", "]", "."]:
                continue
            elif part == "length":
                result = len(result)
            elif part in ["all", "any"]:
                if i + 1 == len(parsed):  # last item
                    results = [
                        self._match_values(item, operator, rhs) for item in result
                    ]
                elif len(parsed) > i + 1 and parsed[i + 1] == ".":
                    lhs_rest = "".join(parsed[i + 2 :])
                    results = [
                        self._match_single(item, lhs_rest, operator, rhs)
                        for item in result
                    ]
                else:
                    raise ValueError(f"'any' cannot be followed by {parsed[i + 1]}")
                try:
                    if part == "all":
                        return results and all(results)  # because all([]) is True
                    else:
                        return any(results)
                except TypeError:
                    return False
            else:
                try:
                    result = result[part]
                except (KeyError, IndexError):
                    return False
            if result is None:
                return False
        return self._match_values(result, operator, rhs)

    def _match_values(self, result, operator: str = "", rhs: str = "") -> bool:
        """Match two values."""
        if not operator:
            return bool(result)
        if isinstance(rhs, str):
            if rhs.isdigit():
                rhs = int(rhs)
            else:
                rhs = rhs.strip("\"'")
        if operator == "=":
            return result == rhs
        if operator == "!=":
            return result != rhs
        if operator == "~":
            if isinstance(result, str):
                rhs = str(rhs)
            return rhs in result
        if operator == "!~":
            if isinstance(result, str):
                rhs = str(rhs)
            return rhs not in result
        try:
            if operator == "<":
                return result < rhs
            if operator == ">":
                return result > rhs
            if operator == "<=":
                return result <= rhs
            if operator == ">=":
                return result >= rhs
        except TypeError:
            return False

    def iter_objects(self, db: DbReadBase) -> Generator[PrimaryObject, None, None]:
        """Iterate over primary objects in a Gramps database."""
        for object_name, objects_name in GRAMPS_OBJECT_NAMES.items():
            iter_method = getattr(db, f"iter_{objects_name}")
            for obj in iter_method():
                obj_dict = json.loads(to_json(obj))
                obj_dict["type"] = obj_dict["_class"].lower()
                if self.match(obj_dict):
                    yield obj
